class TaskPredicate(Exception):
    """Base class for task-related semi-predicates."""


class Retry(TaskPredicate):
    #: Optional message
    message = None

    #: Exception (if any) that caused the retry to happen.
    exc = None

    #: Time of retry (ETA) in seconds
    when = None

    def __init__(self, exc=None, when=None, **kwargs):
        self.when = when
        super().__init__(self, exc, when, **kwargs)


class WorkerStopSignal(Exception):
    pass



from datetime import datetime, date
import uuid
import decimal
from typing import List, Dict, Any, Optional
import msgpack

from .utils import parse_iso8601, to_iso_format


def encode_non_standard_msgpack(obj):
    if isinstance(obj, (decimal.Decimal, uuid.UUID)):
        return str(obj)
    if isinstance(obj, (datetime, date)):
        if not isinstance(obj, datetime):
            obj = datetime(obj.year, obj.month, obj.day, 0, 0, 0, 0)
        r = obj.isoformat()
        if r.endswith("+00:00"):
            r = r[:-6] + "Z"
        return r
    return obj


def msgpack_dumps(s, **kwargs):
    return msgpack.packb(s, default=encode_non_standard_msgpack, use_bin_type=True)


class Message:
    __slots__ = (
        "ack_id",
        "body",
        "queue",
        "receive_count",
    )

    def __init__(
        self, ack_id: str, body: bytes, queue: str = "", receive_count: int = 1
    ):
        self.ack_id = ack_id
        self.body = body
        self.queue = queue
        self.receive_count = receive_count

    def decode_task_message(self):
        messages = msgpack.unpackb(self.body, raw=False)
        eta = messages.get("eta", None)
        expires = messages.get("expires", None)
        if eta and isinstance(eta, str):
            messages["eta"] = parse_iso8601(eta)
        if expires and isinstance(expires, str):
            messages["expires"] = parse_iso8601(expires)
        return TaskMessage(**messages)


class TaskMessage:
    __slots__ = ("id", "task", "args", "kwargs", "retries", "eta", "expires")

    def __init__(
        self,
        id: str,
        task: str,
        args: List[Any] = [],
        kwargs: Dict[str, Any] = {},
        retries: int = 3,
        eta: Optional[datetime] = None,
        expires: Optional[datetime] = None,
    ):
        self.id = id
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.retries = retries
        self.eta = eta
        self.expires = expires

    def encode(self) -> bytes:
        return msgpack_dumps(
            {
                "id": self.id,
                "task": self.task,
                "args": self.args,
                "kwargs": self.kwargs,
                "retries": self.retries,
                "eta": to_iso_format(self.eta) if self.eta else None,
                "expires": to_iso_format(self.expires) if self.expires else None,
            }
        )



import re
from datetime import datetime
from pytz import FixedOffset, utc

ISO8601_REGEX = re.compile(
    r"(?P<year>[0-9]{4})(-(?P<month>[0-9]{1,2})(-(?P<day>[0-9]{1,2})"
    r"((?P<separator>.)(?P<hour>[0-9]{2}):(?P<minute>[0-9]{2})"
    r"(:(?P<second>[0-9]{2})(\.(?P<fraction>[0-9]+))?)?"
    r"(?P<timezone>Z|(([-+])([0-9]{2}):([0-9]{2})))?)?)?)?"
)
TIMEZONE_REGEX = re.compile(
    r"(?P<prefix>[+-])(?P<hours>[0-9]{2}).(?P<minutes>[0-9]{2})"
)


def parse_iso8601(datestr: str) -> datetime:
    m = ISO8601_REGEX.match(datestr)
    if not m:
        raise ValueError(f"unable to parse date string {datestr}")
    groups = m.groupdict()
    tz = groups["timezone"]
    if tz == "Z":
        tz = FixedOffset(0)
    elif tz:
        m = TIMEZONE_REGEX.match(tz)
        prefix, hours, minutes = m.groups()
        hours, minutes = int(hours), int(minutes)
        if prefix == "-":
            hours = -hours
            minutes = -minutes
        tz = FixedOffset(minutes + hours * 60)
    return datetime(
        int(groups["year"]),
        int(groups["month"]),
        int(groups["day"]),
        int(groups["hour"] or 0),
        int(groups["minute"] or 0),
        int(groups["second"] or 0),
        int(groups["fraction"] or 0),
        tz,
    )


def utcnow():
    t = datetime.utcnow()
    return t.replace(tzinfo=utc)


def to_iso_format(d: datetime) -> str:
    r = d.isoformat()

    if r.endswith("+00:00"):
        r = r[:-6] + "Z"

    return r


def countdown(num: int, step: int):
    while num > 0:
        if num > step:
            yield step
        else:
            yield num
        num -= step


def try_to_int(value: str, default: int = 0) -> int:
    try:
        return int(value)
    except ValueError:
        return default



import re
from typing import Optional, Iterable


class Rule:
    def __init__(self, pattern: str, queue: str):
        self.pattern = re.compile(pattern)
        self.queue = queue

    def is_match(self, task_name: str) -> bool:
        return self.pattern.match(task_name)


def route(task_name: str, rules: Iterable[Rule]) -> Optional[str]:
    for rule in rules:
        if rule.is_match(task_name):
            return rule.queue



from typing import Dict, List, TypedDict, Any, Union
from datetime import datetime
import uuid

from .exceptions import Retry
from .messages import Message, TaskMessage


class TaskInfo(TypedDict):
    args: List[Any]
    kwargs: Dict[str, Any]
    task_id: Union[None, str]
    queue: Union[None, str]
    eta: Union[None, datetime]
    expires: Union[None, datetime]


class class_property:
    def __init__(self, getter=None, setter=None):
        if getter is not None and not isinstance(getter, classmethod):
            getter = classmethod(getter)
        if setter is not None and not isinstance(setter, classmethod):
            setter = classmethod(setter)
        self.__get = getter
        self.__set = setter

        info = getter.__get__(object)  # just need the info attrs.
        self.__doc__ = info.__doc__
        self.__name__ = info.__name__
        self.__module__ = info.__module__

    def __get__(self, obj, type=None):
        if obj and type is None:
            type = obj.__class__
        return self.__get.__get__(obj, type)()

    def __set__(self, obj, value):
        if obj is None:
            return self
        return self.__set.__get__(obj)(value)

    def setter(self, setter):
        return self.__class__(self.__get, setter)


class Task:
    _app = None

    name = None

    max_retries = 3

    __bound__ = False

    def retry(self, exc=None, when=None, **kwargs):
        raise Retry(exc=exc, when=when, **kwargs)

    @classmethod
    def bind(cls, app):
        cls.__bound__ = True
        cls._app = app

        cls.on_bound(app)

        return app

    @classmethod
    def on_bound(cls, app):
        pass

    @classmethod
    def _get_app(cls):
        if not cls.__bound__:
            # The app property's __set__  method is not called
            # if Task.app is set (on the class), so must bind on use.
            cls.bind(cls._app)
        return cls._app

    app = class_property(_get_app, bind)

    async def run(self, *args, **kwargs):
        raise NotImplementedError("Tasks must define the run method.")

    def to_message(
        self,
        args=[],
        kwargs={},
        task_id: str = None,
        queue=None,
        eta=None,
        expires=None,
    ) -> Message:
        if task_id is None:
            task_id = str(uuid.uuid4())
        task_message = TaskMessage(
            task_id, self.name, args, kwargs, self.max_retries, eta, expires
        )
        return Message("", task_message.encode(), queue)

    async def batch(self, batches: List[TaskInfo], queue: str = None):
        messages = [self.to_message(**batch) for batch in batches]
        return await self._app.send_batch(
            messages, queue=queue or self._app.task_route(self)
        )

    async def apply(
        self,
        args=[],
        kwargs={},
        task_id: str = None,
        queue=None,
        eta=None,
        expires=None,
        **opts,
    ):
        return await self._app._send_message(
            self.to_message(
                args=args,
                kwargs=kwargs,
                task_id=task_id,
                queue=queue or self._app.task_route(self),
                eta=eta,
                expires=expires,
            ),
            queue=None,
            **opts,
        )


class CronJob:
    name = None

    expression = "* * * * *"

    async def run(self):
        raise NotImplementedError("CronJob must define the run method.")



import asyncio
import logging
import base64
from datetime import datetime
from croniter import croniter
import signal
import uuid
from typing import Optional, List, Callable, Union, Dict

from .driver import QueueDriver
from .exceptions import Retry
from .tasks import Task, CronJob
from .messages import Message
from .routes import route, Rule
from .utils import try_to_int, utcnow, parse_iso8601


async def run_tab(logger, tab, name):
    logger.info("executing cron tab: %s", name)
    await tab.run()


class App:
    def __init__(
        self,
        driver_factory: Callable[[], QueueDriver],
        task_timeout: int = 300,
        queue_size: int = 100,
        default_queue: str = "default",
        routes: Optional[List[Rule]] = None,
        disable_cron: bool = False,
    ):
        self._task_timeout = task_timeout
        self._queue_size = queue_size
        self.driver_factory = driver_factory
        self._driver: Union[None, QueueDriver] = None
        self.logger = logging.getLogger("samping")
        self._tasks: Dict[str, Task] = {}
        self._cron_tabs: Dict[str, CronJob] = {}
        self._loop = None
        self._queue = None
        self.default_queue = default_queue
        self.routes = routes or None
        self.disable_cron = disable_cron

    def task_route(self, task: Task) -> str:
        queue = route(task.name, self.routes)
        self.logger.info("route %s to queue %s", task.name, queue or self.default_queue)
        return queue or self.default_queue

    @property
    def driver(self):
        if self._driver is None:
            self._driver = self.driver_factory()

        return self._driver

    @property
    def queue(self):
        if self._queue is None:
            self._queue = asyncio.Queue(maxsize=self._queue_size)
        return self._queue

    @property
    def loop(self):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop

    def tab(self, **opts):
        def create_job_cls(fun):
            return self._cron_from_fun(fun, **opts)

        return create_job_cls

    def task(self, **opts):
        def create_task_cls(fun):
            return self._task_from_fun(fun, **opts)

        return create_task_cls

    async def run_matched_tabs(self, current_date: Optional[datetime] = None):
        n = utcnow() if current_date is None else current_date
        tasks = []
        for name, tab in self._cron_tabs.items():
            if croniter.match(tab.expression, n):
                task = asyncio.create_task(
                    asyncio.wait_for(
                        run_tab(self.logger, tab, name), timeout=self._task_timeout
                    ),
                    name=name + "-" + str(uuid.uuid4()),
                )
                tasks.append(task)
        if tasks:
            try:
                done, _ = await asyncio.wait(tasks)
                for task in done:
                    try:
                        await task
                    except Exception as e:
                        self.logger.error(
                            f"cron job {task.get_name()} exited with error: {e}"
                        )
            except Exception as e:
                self.logger.error(f"cron jobs exited with error: {e}")

    def __call__(self, event: dict, context: dict):
        if self.is_queue_event(event):
            messages = self.decode_sqs_events(event["Records"])
            self.logger.debug("receive sqs queue %d", len(messages))
            self.on_receive_message(messages)
        elif self.is_scheduled_event(event):
            self.handle_scheduled_event(event)
        else:
            raise TypeError("Invalid SQS event or EventBridge event")

    def is_scheduled_event(self, event: dict) -> bool:
        return event.get("source", "") == "aws.events" and "time" in event

    def is_queue_event(self, event: dict) -> bool:
        return (
            "Records" in event
            and len(event["Records"]) > 0
            and "body" in event["Records"][0]
        )

    async def _signal_handler(self, signame):
        self.logger.info("receive signal %s! exiting", signame)
        await self.queue.join()
        self.loop.stop()

    async def run_worker(self, queues: str, num_worker: int = 3):
        self.logger.info("starting %d workers processing queue %s", num_worker, queues)
        for signame in ("SIGINT", "SIGTERM"):
            self.loop.add_signal_handler(
                getattr(signal, signame),
                lambda: asyncio.create_task(self._signal_handler(signame)),
            )

        tasks = [self._run_producer(queues), self._run_worker(num_worker)]
        if not self.disable_cron:
            tasks.append(self._run_cron())

        return await asyncio.gather(*tasks)

    async def _run_producer(self, queues: str):
        i = 0
        while True:
            try:
                async for message in self.driver.consume(queues):
                    self.logger.debug(
                        "got a new task message, put it in internal queue"
                    )
                    await self.queue.put(message)
            except Exception:
                i += 1
                self.logger.exception("producer %d exited, make it running again", i)

    async def _run_worker(self, num_worker: int = 3):
        return await asyncio.gather(*[self.worker() for _ in range(num_worker)])

    async def worker(self):
        while True:
            try:
                message = await self.queue.get()
                await self.handle_message(message)
            except Exception:
                self.logger.exception("failed to handle a message")
            finally:
                self.queue.task_done()

    async def _run_cron(self):
        while True:
            await self.run_matched_tabs()
            await asyncio.sleep(60)

    def decode_sqs_events(self, sqs_messages: List[dict]) -> List[Message]:
        messages: List[Message] = []
        for sqs_message in sqs_messages:
            body = base64.standard_b64decode(sqs_message.get("body", ""))
            attributes = sqs_message.get("attributes", {})
            receiptHandle = sqs_message.get("receiptHandle", "")
            queue = sqs_message.get("eventSourceARN", "").split(":")[-1]
            messages.append(
                Message(
                    receiptHandle,
                    body,
                    queue=queue,
                    receive_count=try_to_int(
                        attributes.get("ApproximateReceiveCount", 0)
                    ),
                )
            )
        return messages

    def on_receive_message(self, messages: List[Message]):
        handle_message_instance = self.handle_messages(messages, managed=False)
        handle_message_task = self.loop.create_task(handle_message_instance)
        self.loop.run_until_complete(handle_message_task)

    async def handle_messages(self, messages: List[Message], managed: bool = True):
        await asyncio.gather(
            *[self.handle_message(message, managed=managed) for message in messages]
        )

    async def handle_message(self, message: Message, managed: bool = True):
        task_message = message.decode_task_message()
        queue_name = message.queue or self.default_queue
        if task_message.expires is not None and task_message.expires < utcnow():
            await self.driver.send_ack(message.ack_id, queue_name)
        if task_message.task not in self._tasks:
            await self.driver.send_ack(message.ack_id, queue_name)
            return
        task = self._tasks[task_message.task]
        try:
            self.logger.debug("handle message %s", task.name)
            await asyncio.wait_for(
                task.run(*task_message.args, **task_message.kwargs),
                timeout=self._task_timeout,
            )
            await self.driver.send_ack(message.ack_id, queue_name)
        except Retry as exc:
            if managed:
                if task_message.retries < message.receive_count:
                    await self.driver.send_ack(message.ack_id, queue_name)
                else:
                    await self.driver.send_nack(
                        message.ack_id, queue_name, delay=exc.when or 1
                    )
            else:
                raise exc
        except Exception as exc:
            self.logger.exception("handle task %s returned an error", task_message.task)
            if managed:
                if task_message.retries < message.receive_count:
                    await self.driver.send_ack(message.ack_id, queue_name)
                else:
                    await self.driver.send_nack(message.ack_id, queue_name)
            else:
                raise exc

    async def _send_message(
        self, message: Message, queue: Optional[str] = None, **kwargs
    ):
        return await self.send_batch(
            messages=[message], queue=queue or self.default_queue, **kwargs
        )

    async def send_batch(
        self, messages: List[Message], queue: Optional[str] = None, **kwargs
    ):
        return await self.driver.send_batch(
            queue=queue or self.default_queue, messages=messages, **kwargs
        )

    def handle_scheduled_event(self, event: dict):
        if self.is_scheduled_event(event):
            try:
                event_time = parse_iso8601(event.get("time", ""))
            except ValueError:
                event_time = None

            cron_instance = self.run_matched_tabs(current_date=event_time)
            cron_task = self.loop.create_task(cron_instance)
            self.loop.run_until_complete(cron_task)
        else:
            raise TypeError("not an event bridge event")

    def _task_from_fun(self, fun, base=None, name=None, bind=False, **options) -> Task:
        name = name or self.gen_task_name(fun.__name__, fun.__module__)
        base = base or Task
        if name not in self._tasks:
            run = fun if bind else staticmethod(fun)
            task = type(
                fun.__name__,
                (base,),
                dict(
                    {
                        "app": self,
                        "name": name,
                        "run": run,
                        "_decorated": True,
                        "__doc__": fun.__doc__,
                        "__module__": fun.__module__,
                        "__annotations__": fun.__annotations__,
                        "__wrapped__": fun,
                    },
                    **options,
                ),
            )()
            try:
                task.__qualname__ = fun.__qualname__
            except AttributeError:
                pass

            self._tasks[task.name] = task
            task.bind(self)
        else:
            task = self._tasks[name]

        return task

    def _cron_from_fun(
        self, fun, base=None, name=None, bind=False, **options
    ) -> CronJob:
        name = name or self.gen_task_name(fun.__name__, fun.__module__)
        base = base or CronJob
        if name not in self._cron_tabs:
            run = fun if bind else staticmethod(fun)
            task = type(
                fun.__name__,
                (base,),
                dict(
                    {
                        "app": self,
                        "name": name,
                        "run": run,
                        "_decorated": True,
                        "__doc__": fun.__doc__,
                        "__module__": fun.__module__,
                        "__annotations__": fun.__annotations__,
                        "__wrapped__": fun,
                    },
                    **options,
                ),
            )()
            try:
                task.__qualname__ = fun.__qualname__
            except AttributeError:
                pass
            self._cron_tabs[task.name] = task
        else:
            task = self._cron_tabs[name]
        return task

    def gen_task_name(self, name: str, module_name: str) -> str:
        return ".".join(p for p in (module_name, name) if p)



import base64
import aioboto3
from itertools import zip_longest
from typing import (
    Any,
    AsyncGenerator,
    List,
    Iterable,
    TypeVar,
    Optional,
)
import logging
import asyncio
from botocore.exceptions import ClientError


from ..messages import Message
from ..backoff import Backoff, Exponential
from . import QueueDriver
from ..utils import countdown, try_to_int

T = TypeVar("T")


def chunk(iterable: Iterable[T], n: int) -> Iterable[List[T]]:
    args = [iter(iterable)] * n
    grouper = zip_longest(*args, fillvalue=None)
    for chunk in grouper:
        yield list(filter(None, chunk))


class SQSDriver(QueueDriver):
    def __init__(
        self,
        backoff: Optional[Backoff] = None,
        visibility_timeout: int = 30,
        batch_window: int = 5,
        batch_size: int = 10,
        prefetch_size: int = 50,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        region_name: Optional[str] = None,
        botocore_session: Optional[Any] = None,
        profile_name: Optional[str] = None,
        use_ssl: bool = True,
        verify: Optional[bool] = None,
        endpoint_url: Optional[str] = None,
    ):
        self._backoff = backoff or Exponential()
        self.session = aioboto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
            region_name=region_name,
            botocore_session=botocore_session,
            profile_name=profile_name,
        )
        self.use_ssl = use_ssl
        self.verify = verify
        self.endpoint_url = endpoint_url
        self.visibility_timeout = visibility_timeout
        self.lock = asyncio.Lock()
        self._queues = {}
        self.batch_window = batch_window
        self.batch_size = min(batch_size, 20)
        self.logger = logging.getLogger("samping")
        self._in_flight = 0
        self._prefetch_size = prefetch_size

    def sqs_client(self):
        return self.session.resource(
            "sqs",
            use_ssl=self.use_ssl,
            verify=self.verify,
            endpoint_url=self.endpoint_url,
        )

    async def get_queue(self, sqs: Any, queue: str):
        async with self.lock:
            if queue in self._queues:
                return await sqs.Queue(self._queues[queue])
            try:
                sqs_queue = await sqs.get_queue_by_name(QueueName=queue)
            except ClientError as error:
                if (
                    error.response["Error"]["Code"]
                    == "AWS.SimpleQueueService.NonExistentQueue"
                ):
                    # create queue
                    sqs_queue = await sqs.create_queue(
                        QueueName=queue,
                    )
                else:
                    raise error
            self._queues[queue] = sqs_queue.url
            return sqs_queue

    async def send_batch(self, queue: str, messages: List[Message], **kwargs):
        return await self._send_batch(queue, messages)

    async def _send_batch(self, queue: str, messages: List[Message]):
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            async with BatchWriter(sqs_queue) as batch:
                for id, message in enumerate(messages):
                    body = base64.standard_b64encode(message.body).decode()
                    entry = {
                        "Id": str(id),
                        "MessageBody": body,
                    }
                    await batch.send_message(entry)

    async def _send_batch_chunk(self, queue: str, messages: List[Message]):
        async with self.sqs_client() as sqs:
            entries = []
            for id, message in enumerate(messages):
                body = base64.standard_b64encode(message.body).decode()
                entry = {
                    "Id": str(id),
                    "MessageBody": body,
                }
                entries.append(entry)
            sqs_queue = await self.get_queue(sqs, queue)
            response = await sqs_queue.send_messages(Entries=entries)
            if len(response.get("Failed", [])) > 0:
                first = response["Failed"][0]
                raise Exception(
                    "sqs send_batch: failed for %d message(s): %s"
                    % (len(response["Failed"]), first.get("Message", ""))
                )

    async def _fetch_messages_on(self, queues: List[str]):
        async with self.sqs_client() as sqs:
            messages = []
            for queue in queues:
                sqs_queue = await self.get_queue(sqs, queue)
                if len(messages) >= self.batch_size:
                    break
                for max_num in countdown(
                    max(self._prefetch_size - self._in_flight, 10),
                    min(self.batch_size, 10),
                ):
                    current_sqs_messages = await sqs_queue.receive_messages(
                        AttributeNames=["All"],
                        MaxNumberOfMessages=max_num,
                        MessageAttributeNames=["All"],
                        VisibilityTimeout=self.visibility_timeout,
                        WaitTimeSeconds=self.batch_window,
                    )
                    if not current_sqs_messages:
                        break
                    current_messages = await asyncio.gather(
                        *[
                            self.decode_message(message, queue)
                            for message in current_sqs_messages
                        ]
                    )
                    messages.extend(current_messages)
            return messages

    async def consume(self, queues: str) -> AsyncGenerator[Message, None]:
        queue_names = list(filter(lambda x: x, queues.split(",")))
        while True:
            next_wait = 0
            if self._in_flight < self._prefetch_size:
                messages = await self._fetch_messages_on(queue_names)
                if messages:
                    self._backoff.reset()
                    self._in_flight += len(messages)
                    for message in messages:
                        yield message
                else:
                    next_wait = self._backoff.next_backoff().total_seconds()
                    self.logger.info("sqs: got empty messages")
            else:
                next_wait = self._backoff.next_backoff().total_seconds()
                self.logger.info(
                    f"waiting in {self._in_flight} flight message to be processed (max: {self._prefetch_size})"
                )

            if next_wait == 0:
                self._backoff.reset()
                next_wait = self.batch_window
            self.logger.info(
                f"sqs: sleeping for {next_wait} seconds before polling again"
            )
            await asyncio.sleep(next_wait)

    async def send_ack(self, ack_id: str, queue: str = None):
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            message = await sqs_queue.Message(ack_id)
            await message.delete()
            self._in_flight = max(0, self._in_flight - 1)

    async def send_nack(self, ack_id: str, queue: str = None, delay: int = 1):
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            message = await sqs_queue.Message(ack_id)
            await message.change_visibility(VisibilityTimeout=delay)
            self._in_flight = max(0, self._in_flight - 1)

    async def decode_message(self, sqs_message, queue: str) -> Message:
        bodyStr = await sqs_message.body
        body = base64.standard_b64decode(bodyStr)
        attributes = await sqs_message.attributes
        return Message(
            sqs_message.receipt_handle,
            body,
            queue,
            try_to_int(attributes.get("ApproximateReceiveCount", 1), 1),
        )

    async def size(self, queue: str) -> int:
        async with self.sqs_client() as sqs:
            sqs_queue = await self.get_queue(sqs, queue)
            attributes = await sqs_queue.attributes
            return attributes.get("ApproximateNumberOfMessages", 0)


class BatchWriter:
    def __init__(self, queue, flush_amount=10, flush_size=250000):
        self._queue = queue
        self._messages = []
        self._current_size = 0
        self._flush_size = flush_size
        self._flush_amount = flush_amount
        self._lock = asyncio.Lock()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._messages:
            await self._flush()

    def get_size(self, message):
        body = message.get("MessageBody", "").encode("utf-8")
        return len(body)

    async def send_message(self, message):
        async with self._lock:
            current_size = self._current_size + self.get_size(message)
            if current_size > self._flush_size:
                await self._flush()
                self._messages.append(message)
                self._current_size += self.get_size(message)
                if len(self._messages) >= self._flush_amount:
                    await self._flush()
            else:
                self._current_size = current_size
                self._messages.append(message)
                if len(self._messages) >= self._flush_amount:
                    await self._flush()

    async def _flush(self):
        response = await self._queue.send_messages(Entries=self._messages)
        self._messages = []
        if len(response.get("Failed", [])) > 0:
            first = response["Failed"][0]
            raise Exception(
                "sqs send_batch: failed for %d message(s): %s"
                % (len(response["Failed"]), first.get("Message", ""))
            )
        self._current_size = 0



import asyncio
import random
from functools import wraps
from datetime import timedelta
from abc import ABC, abstractmethod

from .utils import utcnow


def retry(count: int = 5, on_exception=(Exception,)):
    def wrapper(fun):
        @wraps(fun)
        async def inner(*args, **kwargs):
            retries = 0
            backoff = Exponential()
            while True:
                try:
                    return await fun(*args, **kwargs)
                except on_exception as e:
                    if retries < count:
                        raise e
                    next_wait = backoff.next_backoff().total_seconds()
                    await asyncio.sleep(next_wait)
                    retries += 1

        return inner

    return wrapper


class Backoff(ABC):
    @abstractmethod
    def next_backoff() -> timedelta:
        ...

    @abstractmethod
    def reset() -> None:
        ...


class Exponential(Backoff):
    def __init__(
        self,
        initial_interval=None,
        randomization_factor=None,
        multiplier=None,
        max_interval=None,
        max_elapsed_time=None,
    ):
        self.initial_interval = initial_interval or timedelta(microseconds=500)
        self.randomization_factor = randomization_factor or 0.5
        self.multiplier = multiplier or 1.5
        self.max_interval = max_interval or 60
        self.max_elapsed_time = max_elapsed_time or timedelta(minutes=3)
        self.start_time = None
        self.current_interval = None
        self.reset()

    def reset(self):
        self.current_interval = self.initial_interval
        self.start_time = utcnow()

    def next_backoff(self):
        elapsed = self.get_elapsed_time()
        next_random = get_random_timedelta(
            self.randomization_factor, random.random(), self.current_interval
        )
        self.increment_interval()
        next_elapsed = elapsed + next_random
        if self.max_elapsed_time and next_elapsed > self.max_elapsed_time:
            return timedelta(seconds=0)
        return next_elapsed

    def get_elapsed_time(self) -> timedelta:
        return utcnow() - self.start_time

    def increment_interval(self):
        if (
            self.current_interval.total_seconds()
            >= self.max_interval // self.multiplier
        ):
            self.current_interval = timedelta(seconds=self.max_interval)
        else:
            self.current_interval = timedelta(
                seconds=self.current_interval.total_seconds() * self.multiplier
            )


def get_random_timedelta(
    randomization_factor: float, random_value: float, current_interval: timedelta
) -> timedelta:
    if randomization_factor == 0:
        return current_interval

    delta = randomization_factor * float(current_interval.total_seconds())
    min_interval = current_interval - timedelta(seconds=delta)
    max_interval = current_interval + timedelta(seconds=delta)

    duration = min_interval.total_seconds() + (
        random_value
        * ((max_interval - min_interval + timedelta(seconds=1)).total_seconds())
    )

    return timedelta(seconds=duration)



import datetime
from samping.messages import Message, TaskMessage


def test_encode_task_message():
    task_message = TaskMessage("id", "test_encode")
    body = task_message.encode()
    message = Message("ack_id", body)
    assert message.queue == ""
    assert message.ack_id == "ack_id"
    task_message_2 = message.decode_task_message()

    assert task_message.id == task_message_2.id
    assert task_message.task == task_message_2.task


def test_encode_message_with_eta():
    task_message = TaskMessage("id", "test_encode", eta=datetime.datetime.now())
    body = task_message.encode()
    message = Message("ack_id", body)
    task_message_2 = message.decode_task_message()

    assert isinstance(task_message_2.eta, datetime.datetime)



from typing import List, Protocol, AsyncGenerator
from abc import abstractmethod

from ..messages import Message


class QueueDriver(Protocol):
    @abstractmethod
    async def send_batch(self, queue: str, messages: List[Message], **kwargs) -> None:
        ...

    @abstractmethod
    async def consume(self, queues: str) -> AsyncGenerator[Message, None]:
        ...

    @abstractmethod
    async def send_ack(self, ack_id: str, queue: str = None) -> None:
        ...

    @abstractmethod
    async def send_nack(self, ack_id: str, queue: str = None, delay: int = 1) -> None:
        ...



from samping.routes import Rule, route


rules = [
    Rule("backend_*", "backend"),
    Rule("ml_*", "ml"),
]


def test_task_routes():
    queue = route("ml_predict", rules)
    assert queue == "ml"


def test_task_no_match():
    queue = route("ml_predict", [])
    assert queue is None


def test_route_exact_match():
    rules = [
        Rule("backend_*", "backend"),
        Rule("ml_predict", "ml_predict"),  # order is important
        Rule("ml_*", "ml"),
    ]
    queue = route("ml_predict", rules)
    assert queue == "ml_predict"
